﻿-- Automation 4 script
--[[
Данный скрипт покадрово генерирует путём интерполяции промежуточные положения
двигающегося по экрану объекта между опорными узлами, пытаясь отрисовать движение с ускорением.
]]

script_name = "Auto tracking"
script_description = "Покадрово рисует положение двигающегося объекта по опорным точкам"
script_author = "cyril51rus"
script_version = "1.3"


-- текст используемой в макросе функции
function Lagrange(x, n_n, n_x, n_y)
    -- Интерполяционный полином Лагранжа
    -- считается значение (res) полинома в точке х (в нашем случае - в момент начала кадра)
    -- n_n - количество узлов интерполяции
    -- n_x[j] - узлы интерполяции (время)
    -- n_y[j] - узловые значения (координата)
    local i, j, F
    local res = 0
    for i = 1 , n_n do
        F = 1
        for j = 1, n_n do
            if i~=j then F = F*(x-n_x[j])/(n_x[i]-n_x[j]) end
        end
        res = res + F*n_y[i];
    end
    return res
-- Данная функция взята из интернета. Она ОЧЕНЬ не оптимальна, т.к.
-- одни и те же коэффициенты полинома рассчитываются в ней при каждом обращении.
-- Но переписывать не буду - влом разбираться. Всё равно и так работает нормально.
end -- конец текста функции


-- начало тела самого нашего макроса
function auto_tracking(subtitles, selected_lines, active_line) -- selected_lines и active_line здесь не используются

    node_style = "Tracking" -- новый стиль всех наших добавленных строк с трекингом
    local num_of_nodes = 0 -- количество заданных пользователем узлов (от 2 до 9)
    local node_index = {} -- массив индексов (номеров строк в текущем файле) узловых строк
    local node_x = {} -- массив X-координат узлов
    local node_y = {} -- массив Y-координата узлов
    local node_start_time = {} -- массив начальных времён узлов
    local node_end_time = {} -- массив начальных времён узлов
    local node_fs = {} -- массив величины шрифта текста в узлах
    local node_frz = {} -- массив углов поворота текста в узлах
    local node_fadein = {} -- массив параметров появления текста (вообще-то он не нужен, достаточно только первого узла!)
    local node_fadeout = {} -- массив параметров затухания (тоже не нужен, достаточно только первого узла!)
    local node_frame = {} -- массив номеров кадров начала узлов
    local current_line -- временное хранилище текущей строки субтитра
    local position = nil -- указатель позиции
    local index = 0 -- просто счётчик
    local j = 0 -- ещё один счётчик

    -- перебираем все строки саба и ищем строки с узлами, т.е с фиктивными тегами "\node01" и т.д.
    for index = 1, #subtitles do
        -- обрабатывать будем только в строки с сабами, причём комменты отбрасываются!
        if subtitles[index].class == "dialogue" and subtitles[index].comment ~= true then
            -- пример искомой строки с узлом: {\b1\fs24\node01\pos(267.2,187.2)}Строка{\b0}
            position = string.find(subtitles[index].text,"\\node0%d",1,false)
                -- Achtung! Проверок не делаем: считаем, что строки с опорными узлами всегда правильно оформлены
                -- и всегда идут по возрастанию времени и индекса, в количестве от 2 до 9 строк!
            if position~=nil then -- тег узла найден
                current_line = subtitles[index]
                -- увеличиваем счётчик узлов
                num_of_nodes = num_of_nodes + 1
                -- запоминаем номер строки с узловой точкой (от начала файла субтитров!)
                node_index[num_of_nodes] = index
                -- запоминаем время показа узловых строк
                node_start_time[num_of_nodes] = current_line.start_time -- (в миллисекундах!)
                node_end_time[num_of_nodes] = current_line.end_time
                -- запоминаем номер кадра узла
                node_frame[num_of_nodes] = aegisub.frame_from_ms(current_line.start_time)
                -- Важно: начальный и последующий узловой кадр нужно брать по начальному времени,
                -- а последний кадр - по конечному времени показа надписи!
                -- (*т.е. последний node_frame после окончания перебора потом нужно будет исправить!)
                -- считываем координаты из узловых кадров (Achtung! Проверок тут нет!)
                node_x[num_of_nodes] = tonumber( string.match(current_line.text,"\\pos%((%-?%d+%.?%d*),%-?%d+%.?%d*%)",1) )
                node_y[num_of_nodes] = tonumber( string.match(current_line.text,"\\pos%(%-?%d+%.?%d*%,(%-?%d+%.?%d*)%)",1) )
                -- считываем размер шрифта из узловых кадров (Achtung! Проверок тут нет!)
                node_fs[num_of_nodes] = tonumber( string.match(current_line.text,"\\fs(%d+%.?%d*)[\\%}]+",1) )
                -- (Внимание! Размер шрифта здесь можно задать дробный, но это справедливо только для данного скрипта!)
                -- аналогичным образом считываем и угол поворота текста (Achtung! Проверок тут нет!)
                node_frz[num_of_nodes] = tonumber( string.match(current_line.text,"\\frz?(%-?%d+%.?%d*)[\\%}]+",1) )
                -- аналогичным образом считываем параметры тэга \fad, если он есть (Achtung! Проверок тут нет!)
                node_fadein[num_of_nodes] = tonumber( string.match(current_line.text,"\\fad%((%d+),%d+%)",1) )
                node_fadeout[num_of_nodes] = tonumber( string.match(current_line.text,"\\fad%(%d+,(%d+)%)",1) )
                -- Важно! Если какой-либо из параметров выше в пользовательской строке был пропущен,
                -- то в массив запишется nil!

                -- далее помечаем копию текущей строки, как комментарий
                current_line.comment = true
                -- и перезаписываем её в текущем файле субтитров
                subtitles[index] = current_line
            end -- разборка тегов текущего узла закончена
        end -- конец обработки строки с сабами
    end -- конец поиска опорных узлов
    -- Исправляем последний node_frame (*выставляем его по конечному времени показа последнего узла)
    node_frame[num_of_nodes] = aegisub.frame_from_ms(node_end_time[num_of_nodes])

    -- время в крайних узловых точках также выравняем по границам кадров (а нужно ли это?)
    node_start_time[1] = aegisub.ms_from_frame(node_frame[1])
    node_end_time[num_of_nodes] = aegisub.ms_from_frame(node_frame[num_of_nodes])

    -- проверим количество заданных пользователем узлов
    if num_of_nodes<3 then
	aegisub.progress.title("Error")
	aegisub.debug.out("Меньше трёх узлов задавать нет смысла!\n")
	aegisub.cancel()
    end
    -- (Achtung! Проверки на правильность следования узлов у нас здесь нет!)

    -- Проверим, во всех ли узлах пользователь задал размеры шрифта
    for j=2, num_of_nodes do
        if node_fs[1]~=nil then -- если самый первый размер шрифта (в первом узле) был задан
            -- то проверим и текущий узел
            if node_fs[j]==nil then -- если размер шрифта в текующем узле не задан
                node_fs[j] = node_fs[j-1] -- то присвоим ему предыдущее значение
            end
        else -- если же самый первый размер шрифта был не задан
            -- то не будем учитывать и все остальные значения размера
            node_fs[j] = nil
        end
    end -- конец проверки наличия значений величины шрифта в узлах

    local fadein_speed = 0 -- на тот случай если в тэге \fad начальное значение задано =0
    local fadeout_speed = 0 -- на тот случай если в тэге \fad конечное значение задано =0
    -- Проверим, задал ли пользователь параметры тэга \fad (проверяем только первый узел)
    if node_fadein[1]~=nil and node_fadeout[1]~=nil then -- оба параметра заданы
        -- проверим суммарную длительность: если она больше чем сам трекинг - игнорируем её
        if (node_fadein[1]+node_fadeout[1])<=(node_end_time[num_of_nodes]-node_start_time[1]) then
                if node_fadein[1]~=0 then fadein_speed = 255/node_fadein[1] end
                if node_fadeout[1]~=0 then fadeout_speed = 255/node_fadeout[1] end
        end
    end

    -- Проверим, во всех ли узлах пользователь задал угол поворота текста
    for j=2, num_of_nodes do
        if node_frz[1]~=nil then -- если самый первый угол поворота (в первом узле) был задан
            -- то проверим и текущий узел
            if node_frz[j]==nil then -- если угол поворота в текущем узле не задан
                node_frz[j] = node_frz[j-1] -- то присвоим ему предыдущее значение
            end
        else -- если же самый первый угол не был задан
            -- то не будем учитывать и все остальные углы
            node_frz[j] = nil
        end
    end -- конец проверки наличия значений угла в узлах

    -- Все заданные в узлах углы нужно обязательно проверить на перескок через 0, т.к. угол поворота должен всегда меняться гладко!
    -- (считаем, что текст не будет вертеться более чем на 360 градусов)
    if node_frz[1]~=nil then
        -- Первым делом проверим перескок на первом шаге:
        if (node_frz[1]>270 and node_frz[1]<360) and (node_frz[2]>0 and node_frz[2]<90) then
            -- перескок из 4 четверти в 1 на первом шаге (т.е. из минуса в плюс против ч.с.)
            node_frz[1] = node_frz[1] - 360
        end
        if (node_frz[1]>0 and node_frz[1]<90) and (node_frz[2]>270 and node_frz[2]<360) then
            -- перескок из 1 четверти в 4 на первом шаге (т.е. из плюса в минус по ч.с.)
            node_frz[2] = node_frz[2] - 360
        end
        -- Далее проверяем перескоки на последующих шагах:
        for j=2, num_of_nodes do
            if (node_frz[j-1]>270 and node_frz[j-1]<360) and (node_frz[j]>0 and node_frz[j]<90) then
                -- перескок из 4 четверти в 1 (т.е. из минуса в плюс против ч.с.)
                node_frz[j] = node_frz[j] + 360
            end
            if (node_frz[j-1]>0 and node_frz[j-1]<90) and (node_frz[j]>270 and node_frz[j]<360) then
                -- перескок из 1 четверти в 4 на первом шаге (т.е. из плюса в минус по ч.с.)
                node_frz[j] = node_frz[j] - 360
            end
        end
    -- (Achtung! Что-то мне эти проверки совсем не нравятся! Надо бы подумать получше.)
    end -- конец проверки перескока угла

    -- всего кадров в трекинге, включая узловые точки, будет
    local total_num_of_frames = node_frame[num_of_nodes] - node_frame[1] + 1

    local new_x = {} -- массив X-координат новых точек
    local new_y = {} -- массив Y-координат новых точек
    local new_start_time = {} -- массив начальных времён новых точек
    local new_end_time = {} -- массив конечных времён новых точек
    local new_frame = {} -- массив номеров кадров новых точек
    local new_fs = {} -- массив размеров шрифта в новых точках
    local new_frz = {} -- массив углов поворота текста в новых точках
    local new_alpha = {} -- массив значений тэга прозрачности \alpha&HFF&, рассчитанного на основе значений \fad

    -- далее наполняем новые массивы расчетными параметрами новых кадров
    for index = 1, total_num_of_frames do
        -- определяем номер текущего расчётного кадра
        new_frame[index] = node_frame[1] + index - 1
        -- определяем время показа текущего субтитра (кадра)
        -- (время субтитров с узловыми точками здесь тоже будет выровнено по границам кадров)
        new_start_time[index] = aegisub.ms_from_frame(new_frame[index])
        new_end_time[index] = aegisub.ms_from_frame(new_frame[index]+1)
        -- (Achtung! Я надеюсь, что мы здесь никогда не выйдем за последний кадр, поэтому не проверяемся!)

        -- устанавливаем признак несовпадения (он же - номер совпавшего узла)
        position = 0 -- ="не совпадает"
        -- проверяем: совпадает ли номер текущего расчётного кадра с узловыми
        for j = 1, num_of_nodes do
            if new_frame[index]==node_frame[j] then
                -- если кадр совпадает, выставляем соответствующий признак совпадения
                position = j -- (т.е. запоминаем номер совпавшего узла)
                -- и выходим их цикла
                break
            end
        end -- конец проверки совпадения текущего кадра с узловыми точками

        -- вычисляем параметры промежуточного кадра
        if position>0 then -- т.е. если кадр совпадает с узловой точкой
            -- координаты перемещаемого текста берём из узловой точки "как есть"
            -- (округлять координаты будем не сейчас, а после завершения всех расчётов!)
            new_x[index] = node_x[position]
            new_y[index] = node_y[position]
            -- размер шрифта перемещаемого текста тоже берём из узла "как есть" (в т.ч. и nil!)
            new_fs[index] = node_fs[position]
            -- угол поворота текста тоже берём из узла "как есть" (в т.ч. и nil!)
            new_frz[index] = node_frz[position]
            -- параметры узла скопированы; возвращаем на место признак несовпадения
            position = 0
        else -- а если кадр не совпадает с узлом (т.е. рассчитываем параметры промежуточного кадра)
            -- определяем предположительные координаты перемещаемого текста расчётным способом
            new_x[index] = Lagrange(new_start_time[index], num_of_nodes, node_start_time, node_x)
            new_y[index] = Lagrange(new_start_time[index], num_of_nodes, node_start_time, node_y)
            -- (округлять будем не сейчас, а после завершения всех расчётов!)
            -- теперь разберёмся (без лишнего педантизма) с размером шрифта
            if node_fs[1]~=nil then -- считаем, что пользователь задал размер шрифта во всех узлах
                -- т.е. мы предполагаем, что расчёт шрифта в промежуточных точках требуется
                if node_fs[1]==node_fs[2] then -- размер шрифта по ходу движения не меняется
                    -- везде ставим размер из первого узла:
                    new_fs[index] = node_fs[1]
                else -- а если размер шрифта меняется
                    -- определяем предположительный размер шрифта перемещаемого текста расчётным способом
                    new_fs[index] = Lagrange(new_start_time[index], num_of_nodes, node_start_time, node_fs)
                    -- (его тоже будем округлять не сейчас, а после завершения всех расчётов!)
                end
            else -- пользователь проставил размер шрифта не во всех узлах
                new_fs[index] = nil
            end -- конец разборки со шрифтами
            -- теперь разберёмся (тоже без лишнего педантизма) с углом поворота текста
            if node_frz[1]~=nil then -- считаем, что пользователь задал угол во всех узлах
                -- т.е. мы предполагаем, что это значит, что расчёт угла в промежуточных точках требуется
                if node_frz[1]==node_frz[2] then -- считаем, что угол по ходу движения не меняется
                    new_frz[index] = node_frz[1]
                else -- а если угол меняется, то
                    -- определяем предположительный размер угла расчётным способом
                    new_frz[index] = Lagrange(new_start_time[index], num_of_nodes, node_start_time, node_frz)
                    -- (его тоже будем округлять не сейчас, а после завершения всех расчётов!)
                end
            else -- пользователь проставил угол не во всех узлах
                new_frz[index] = nil
            end -- конец разборки с углами
        end -- конец вычисления параметров промежуточного кадра

        -- теперь разберёмся с параметрами тэга \alpha (если был задан \fad)
        new_alpha[index] = nil -- Помни! alpha здесь будет десятичная!
        if node_fadein[1]~=nil and node_fadeout[1]~=nil then
        --aegisub.debug.out(new_start_time[index].."-"..new_end_time[index].."=="..node_start_time[1].."-"..node_end_time[1].."\n")
            if fadein_speed~=0 and new_start_time[index]<(node_start_time[1]+node_fadein[1]) then
                new_alpha[index] = math.floor( 255.5 - fadein_speed*(new_end_time[index]-node_start_time[1]) )
                if new_alpha[index]<0 then new_alpha[index]=0 end -- можно сделать и nil, чтобы не выводилось
            end
            if fadeout_speed~=0 and new_start_time[index]>=(node_end_time[num_of_nodes]-node_fadeout[1]) then
                new_alpha[index] = math.floor( 0.5 + fadeout_speed*(new_end_time[index]-node_end_time[num_of_nodes]+node_fadeout[1]) )
                if new_alpha[index]>255 then new_alpha[index]=255 end
            end
        end -- конец разборки с прозрачностью

    end -- конец цикла наполнения новых массивов рассчётными точками


-- Блок "оптимизации". Если не хотите, чтобы он выкидывал "лишние" кадры из трекинга,
-- [[ то просто закомментируйте его.
    local delta_x = 0 -- горизонтальное смещение по сравнению с предыдущим кадром
    local delta_y = 0 -- вертикальное смещение по сравнению с предыдущим кадром
    local delta_fs = 0 -- увеличение шрифта по сравнению с предыдущим кадром
    local delta_frz = 0 -- увеличение шрифта по сравнению с предыдущим кадром
    local previous_x = 100000 -- (переменные для хранения предыдущих значений параметров)
    local previous_y = 100000 -- (их начальные значения специально сделаны нереальными)
    local previous_fs = 0 -- (чтобы при первом сравнении не возникло совпадения)
    local previous_frz = 10000
    local previous_index = 100000 -- переменная для хранения номера строки с предыдущим титром
    -- здесь будет происходить оптимизация новых строк
    for index = 1, total_num_of_frames do
        -- вычисляем смещение текста между кадрами
        delta_x = math.abs(new_x[index] - previous_x)
        delta_y = math.abs(new_y[index] - previous_y)
        -- вычисляем изменение размера шрифта между кадрами
        if node_fs[1]~=nil then
            delta_fs = math.abs( new_fs[index] - previous_fs )
        end
        -- вычисляем изменение угла между кадрами
        if node_frz[1]~=nil then
            delta_frz = math.abs(new_frz[index] - previous_frz)
        end
        -- смотрим, насколько сместился/увеличился текст по сравнению с предыдущим кадром
        -- (Интересно: а каковы оптимальные величины смещений, чтобы ничего не дрыгалось?)
        if delta_x<1 and delta_y<1 and delta_fs<1 and delta_frz<1 then -- кадр сместился меньше чем на пиксель
            -- значит предыдущий кадр можно показать чуть подольше
            new_end_time[previous_index] = new_end_time[index] -- (на первом шаге не выполнится)
            -- а текущий кадр можно пропустить, для этого поставим в нём специальную отметку
            new_end_time[index] = nil -- (отличать пропускаемые кадры будем потом по этому признаку)
        else -- кадр изменился значительно и ничего пропустить не получится, следовательно
            -- запоминаем параметры текущего кадра для их сравнения с последующим
            previous_x = new_x[index]
            previous_y = new_y[index]
            previous_fs = new_fs[index]
            previous_frz = new_frz[index]
            previous_index = index
        end -- конец сравнения дельты c 1
    end
    -- конец оптимизации новых строк
-- ]]
-- Конец блока "оптимизации".


    local new_value = "" -- чистая строка для генерации тега \pos()
    -- писать туда в теги будем только целые числа, поэтому подготовим место для округлённых чисел:
    local rounded_x = 0 -- чистая переменная для округлённых X
    local rounded_y = 0 -- чистая переменная для округлённых Y
    local rounded_fs = 0 -- чистая переменная для округлённых fs
    local rounded_frz = 0 -- чистая переменная для округлённых frz
    -- сам перемещаемый текст возьмём из титра в первом узловом кадре
    position = node_index[1] -- индекс первой узловой строки в текущем файле субтитров
    current_line = subtitles[position] -- текст двигающейся строки будем править здесь
    -- сперва уберём оттуда наш фиктивный тэг
    current_line.text = string.gsub(current_line.text,"\\node0%d","")
    -- и отключим ранее проставленный там нами признак комментария
    current_line.comment = false
    -- убираем из текущей строки \fad (если он был туда добавлен)
    current_line.text = string.gsub(current_line.text,"\\fad%b()","")
    -- также назначим ей новый стиль, отличающий строки с трекингом от всех остальных строк
    current_line.style = node_style
    -- для вывода комментов заготовим отдельную строку!
    current_line2 = subtitles[position]
    current_line2.comment = true
    current_line2.text = "Ниже идут расчётные значения положения текста в кадре. При необходимости, корректируйте значения \\pos вручную."
    -- время первого коммента возьмём из первого узла (просто так, для порядка)
    current_line2.start_time = node_start_time[1]
    current_line2.end_time = node_end_time[1]
    -- Первым делом добавляем строку с первым комментарием в конец текущего файла субтитров
    subtitles[0] = current_line2

    -- Потом начинаем последовательную генерацию новых строк с промежуточными точками и добавление их в конец скрипта
    for index = 1, total_num_of_frames do
        -- выводить будем только "оптимизированные" (непомеченные nil'ом) строки!
        -- проверяем строки на наличие пометки
        if new_end_time[index]~=nil then
            -- определяем начальное и конечное время нового субтитра (кадра)
            -- длительность показа узловых точек ограничим длительностью одного кадра
            current_line.start_time = new_start_time[index]
            current_line.end_time = new_end_time[index]
            -- округляем до целых чисел рассчётные значения координат
            rounded_x = math.floor(0.5+new_x[index])
            rounded_y = math.floor(0.5+new_y[index])
            -- формируем новый тег \pos с расчётными значениями
            new_value = string.format('\\pos(%d,%d)', rounded_x, rounded_y)
            -- и дописываем к нему тэг прозрачности, если пользователь задал в первом узле \fad
            if node_fadein[1]~=nil and node_fadeout[1]~=nil then
                -- убираем из текущей строки тэг \alpha (если он был туда добавлен на предыдущих этапах)
                -- Achtung! Если пользователь задал его специально, то его тэг будет удалён!
                current_line.text = string.gsub(current_line.text,"\\alpha%b&&","")
                if new_alpha[index]~=nil then
                    new_value = new_value..string.format('\\alpha&H%02X&',new_alpha[index])
                end
            end
            -- а теперь вписываем в строку субтитра новые координаты перемещаемого текста (с добавленным тэгом прозрачности, если задано)
            current_line.text = string.gsub(current_line.text,"\\pos%(%-?%d+%.?%d*%,%-?%d+%.?%d*%)",new_value)

            if node_fs[1]~=nil then
                rounded_fs = math.floor(0.5+new_fs[index])
                -- формируем новый тег \fs с рассчётными значениями
                new_value = string.format('\\fs%d', rounded_fs)
                -- и вписываем в строку субтитра новый размер шрифта перемещаемого текста
                current_line.text = string.gsub(current_line.text,"\\fs%d+%.?%d*",new_value)
            end
            if node_frz[1]~=nil then
                rounded_frz = math.floor(0.5+new_frz[index])
                -- формируем новый тег \frz с рассчётными значениями
                new_value = string.format('\\frz%d', rounded_frz)
                -- и вписываем в строку субтитра новый размер шрифта перемещаемого текста
                current_line.text = string.gsub(current_line.text,"\\frz?%-?%d+%.?%d*",new_value)
            end
            -- а теперь, добавляем новую строку в конец текущего файла субтитров
            subtitles[0] = current_line
        end -- конец проверки строк на наличие пометок
    end -- конец генерации новых кадров
    -- И наконец, добавляем ещё одну, последнюю, новую строку с комментарием
    -- в самый конец текущего файла субтитров
    current_line2.text = "Конец выдачи рассчётных значений."
    -- время последнего коммента возьмём из последнего узла (просто так, для порядка)
    current_line2.start_time = node_start_time[num_of_nodes]
    current_line2.end_time = node_end_time[num_of_nodes]
    subtitles[0] = current_line2

    -- делаем обязательный COMMIT
    aegisub.set_undo_point(script_name)
end
-- конец тела нашего макроса

aegisub.register_macro(script_name, script_description, auto_tracking)

-- конец скрипта
